@@ -21,21 +21,33 @@
 
 using namespace std;
 
-// sync is a proper superset of announce & beacon,
+// chat is a proper superset of announce/beacon,
+// sync is a proper superset of chat,
 // both inbound as well as outbound.
-// in these 2 arrays of strings, sync-specific
-// fields are placed after announce fields.
+// in these 2 arrays of strings, chat-specific fields follow
+// announce fields, and sync-specific follow chat fields.
+// (chat overloads bible for this purpose.)
 // see field_count in ReceiveInternal + TransmitInternal.
 static string inbound_required[] = {
     BSP_APP_NAME,
     BSP_APP_INSTANCE_UUID,
     BSP_APP_USER,
     BSP_MSG_PASSPHRASE,
+    BSP_MSG_SYNC_BIBLEABBREV,
     BSP_MSG_SYNC_DOMAIN,
     BSP_MSG_SYNC_VERSE,
-    BSP_MSG_SYNC_BIBLEABBREV,
     BSP_MSG_SYNC_GROUP
 };
+#define	CHAT_INBOUND_INDEX	4	// index in list above.
+
+static int inbound_required_count[5] =
+{
+	0,				// unused
+	BSP_FIELDS_RECV_ANNOUNCE,
+	BSP_FIELDS_RECV_SYNC,
+	BSP_FIELDS_RECV_ANNOUNCE,	// beacon identical to announce
+	BSP_FIELDS_RECV_CHAT
+};
 
 static string outbound_fill[] = {
     BSP_APP_NAME,
@@ -45,12 +57,24 @@ static string outbound_fill[] = {
     BSP_APP_DEVICE,
     BSP_APP_USER,
     BSP_MSG_PASSPHRASE,
+    BSP_MSG_SYNC_BIBLEABBREV,
     BSP_MSG_SYNC_DOMAIN,
     BSP_MSG_SYNC_GROUP,
-    BSP_MSG_SYNC_BIBLEABBREV,
     BSP_MSG_SYNC_ALTVERSE,
     BSP_MSG_SYNC_VERSE		// last: could go overly long, risk cutoff.
 };
+#define	CHAT_OUTBOUND_INDEX	7	// index in list above.
+
+static int outbound_fill_count[5] =
+{
+	0,				// unused
+	BSP_FIELDS_XMIT_ANNOUNCE,
+	BSP_FIELDS_XMIT_SYNC,
+	BSP_FIELDS_XMIT_ANNOUNCE,	// beacon identical to announce
+	BSP_FIELDS_XMIT_CHAT
+};
+
+static string chat_field = BSP_MSG_CHAT;	// for referential substitution.
 
 // BibleSync class constructor.
 // args identify the user of the class, by application, version, and user.
@@ -368,7 +392,9 @@ int BibleSync::ReceiveInternal()
 				   ? "sync"
 				   : ((bsp.msg_type == BSP_BEACON)
 				      ? "beacon"
-				      : "*???*"))),
+				      : ((bsp.msg_type == BSP_CHAT)
+					 ? "chat"
+					 : "*???*")))),
 		 uuid_dump_string,
 		 bsp.num_packets, bsp.index_packet,
 		 bsp.body);
@@ -378,13 +404,15 @@ int BibleSync::ReceiveInternal()
 	    (*nav_func)('E', EMPTY,
 			EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
 			BSP + _("bad magic"), dump);
-	else if (bsp.version != BSP_PROTOCOL)
+	else if ((bsp.version != BSP_PROTOCOL) && (bsp.version != BSP_OLD_PROTOCOL))
+	    // we are fine with previous v2 protocol that lacks chat messages.
 	    (*nav_func)('E', EMPTY,
 			EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
 			BSP + _("bad protocol version"), dump);
 	else if ((bsp.msg_type != BSP_ANNOUNCE) &&
 		 (bsp.msg_type != BSP_SYNC) &&
-		 (bsp.msg_type != BSP_BEACON))
+		 (bsp.msg_type != BSP_BEACON) &&
+		 (bsp.msg_type != BSP_CHAT))
 	    (*nav_func)('E', EMPTY,
 			EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
 			BSP + _("bad msg type"), dump);
@@ -444,13 +472,16 @@ int BibleSync::ReceiveInternal()
 	    else
 	    {
 		// verify minimum body content.
-		int field_count = ((bsp.msg_type == BSP_SYNC)
-				   ? BSP_FIELDS_RECV_SYNC
-				   : BSP_FIELDS_RECV_ANNOUNCE);	// or beacon.
+		int field_count = inbound_required_count[bsp.msg_type];
 
 		for (int i = 0; i < field_count; ++i)
 		{
-		    if (content.find(inbound_required[i]) == content.end())
+		    string &locator = (((bsp.msg_type == BSP_CHAT) &&
+					(i == CHAT_INBOUND_INDEX))
+				       ? chat_field
+				       : inbound_required[i]);
+
+		    if (content.find(locator) == content.end())
 		    {
 			ok_so_far = false;
 			string info = BSP + _("missing required header ")
@@ -527,7 +558,25 @@ int BibleSync::ReceiveInternal()
 			version = (string)"(version?)";
 		
 		    // generally good, so extract interesting content.
-		    if (bsp.msg_type == BSP_SYNC)
+		    if (bsp.msg_type == BSP_CHAT)
+		    {
+			bible  = content.find(BSP_APP_USER)->second;
+			ref    = source_addr;
+			group  = content.find(BSP_APP_NAME)->second
+			    + " " + version;
+			domain = content.find(BSP_APP_DEVICE)->second;
+			alt    = content.find(BSP_MSG_CHAT)->second;
+
+			info = (string)"chat: "
+			    + content.find(BSP_APP_USER)->second
+			    + " @ " + source_addr;
+
+			cmd = ((passphrase ==
+				content.find(BSP_MSG_PASSPHRASE)->second)
+			       ? 'C'	// chat message
+			       : 'M');	// mismatch
+		    }
+		    else if (bsp.msg_type == BSP_SYNC)
 		    {
 			// regular synchronized navigation
 			bible  = content.find(BSP_MSG_SYNC_BIBLEABBREV)->second;
@@ -745,7 +794,8 @@ BibleSync::TransmitInternal(char message_type,
 
     if ((message_type != BSP_ANNOUNCE) &&
 	(message_type != BSP_SYNC) &&
-	(message_type != BSP_BEACON))
+	(message_type != BSP_BEACON) &&
+	(message_type != BSP_CHAT))
 	return BSP_XMIT_BAD_TYPE;
 
     if ((mode == BSP_MODE_AUDIENCE) &&
@@ -757,18 +807,25 @@ BibleSync::TransmitInternal(char message_type,
     string body = "";
 
     // all name/value pairs.
-    content[BSP_APP_NAME]             = application;
-    content[BSP_APP_VERSION]          = version;
-    content[BSP_APP_USER]             = user;
-    content[BSP_APP_DEVICE]           = device;
-    content[BSP_APP_OS]               = BSP_OS;
-    content[BSP_APP_INSTANCE_UUID]    = uuid_string;
-    content[BSP_MSG_SYNC_BIBLEABBREV] = bible;
-    content[BSP_MSG_SYNC_VERSE]       = ref;
-    content[BSP_MSG_SYNC_ALTVERSE]    = alt;
-    content[BSP_MSG_SYNC_GROUP]       = group;
-    content[BSP_MSG_SYNC_DOMAIN]      = domain;
-    content[BSP_MSG_PASSPHRASE]       = passphrase;
+    content[BSP_APP_NAME]                 = application;
+    content[BSP_APP_VERSION]              = version;
+    content[BSP_APP_USER]                 = user;
+    content[BSP_APP_DEVICE]               = device;
+    content[BSP_APP_OS]                   = BSP_OS;
+    content[BSP_APP_INSTANCE_UUID]        = uuid_string;
+    if (message_type != BSP_CHAT)
+	content[BSP_MSG_SYNC_BIBLEABBREV] = bible;
+    else {
+	content[BSP_MSG_CHAT]	          = bible;	// overload.
+	// innoculate chat content against internal \n.
+	for (char *chase = (char *)bible.c_str(); chase = strchr(chase, '\n'); ++chase)
+	    *chase = '\t';
+    }
+    content[BSP_MSG_SYNC_VERSE]           = ref;
+    content[BSP_MSG_SYNC_ALTVERSE]        = alt;
+    content[BSP_MSG_SYNC_GROUP]           = group;
+    content[BSP_MSG_SYNC_DOMAIN]          = domain;
+    content[BSP_MSG_PASSPHRASE]           = passphrase;
 
     // header.
     bsp.magic = BSP_MAGIC;
@@ -780,13 +837,15 @@ BibleSync::TransmitInternal(char message_type,
     memset((void *)&bsp.reserved, 0, BSP_RES_SIZE);
 
     // body prep.
-    int field_count = ((message_type == BSP_SYNC)
-		       ? BSP_FIELDS_XMIT_SYNC
-		       : BSP_FIELDS_XMIT_ANNOUNCE);	// or beacon.
+    int field_count = outbound_fill_count[message_type];
 
     for (int i = 0; i < field_count; ++i)
     {
-	body += outbound_fill[i] + "=" + content[outbound_fill[i]] + "\n";
+	string &filler = (((message_type == BSP_CHAT) &&
+			   (i == CHAT_OUTBOUND_INDEX))
+			  ? chat_field
+			  : outbound_fill[i]);
+	body += filler + "=" + content[filler] + "\n";
     }
 
     // ship it.
@@ -795,6 +854,7 @@ BibleSync::TransmitInternal(char message_type,
 				 BSP_HEADER_SIZE + body.length());
 
     // force last == newline: attempt to preserve body format when long.
+    // (cuts off excessively long verse references and chat messages.)
     ((unsigned char*)&bsp)[BSP_MAX_SIZE-1] = '\n';
 
     BibleSync_xmit_status retval;
@@ -896,179 +956,62 @@ void BibleSync::clearSpeakers()
 
 #ifdef linux
 
-// routines below imported from the net as workable examples.
 // in order to do multicast setup, we require the address
-// of the interface that has our default route.  code below
-// finds the name of that interface.  then getifaddrs(3)
-// code (taken from its man page) lets us match that name
-// against an entry that has the address we need.
+// of the interface that has our default route.
+// get_default_if_name() reads /proc/net/route to find that interface.
+// then getifaddrs(3) code (taken from its man page) lets us
+// match that name against an entry that has the address we need.
+// this entire methodology is 100 times simpler than the former
+// rtnetlink-driven nightmare
 
-// extra includes needed only for the
-// route & gateway discovery code below.
+// lines in /proc/net/route consist of
+// IFACE \t DESTINATION \t GATEWAY \t FLAGS \t ...
+// DESTINATION is an 8-byte hex value (string), so look for \t00000000\t.
+
+#define	PROC_ROUTE	"/proc/net/route"
 
 #include <net/if.h>
-#include <linux/rtnetlink.h>
+#include <netdb.h>
+#include <ifaddrs.h>
 
-struct route_info
+int BibleSync::get_default_if_name(char *name)
 {
-    struct in_addr dstAddr;
-    struct in_addr srcAddr;
-    struct in_addr gateWay;
-    char ifName[IF_NAMESIZE];
-};
+    int found = 0;
+    char line[256], *field;
+    FILE *proc_route;
 
-int BibleSync::readNlSock(int sockFd,
-			  char *bufPtr,
-			  size_t buf_size,
-			  unsigned int seqNum,
-			  unsigned int pId)
-{
-    struct nlmsghdr *nlHdr;
-    int readLen = 0, msgLen = 0;
-
-    do
-    {
-	if((readLen = recv(sockFd, bufPtr, buf_size - msgLen, 0)) < 0)
-	{
-	    return -1;
-	}
-
-	nlHdr = (struct nlmsghdr *)bufPtr;
-
-	if((NLMSG_OK(nlHdr, readLen) == 0) || (nlHdr->nlmsg_type == NLMSG_ERROR))
-	{
-	    return -1;
-	}
-
-	/* check if last message */
-	if(nlHdr->nlmsg_type == NLMSG_DONE)
-	{
-	    break;
-	}
-	else
-	{
-	    /* else move pointer to buffer appropriately */
-	    bufPtr += readLen;
-	    msgLen += readLen;
-	}
-
-	/* check if multi part message */
-	if((nlHdr->nlmsg_flags & NLM_F_MULTI) == 0)
-	{
-	    /* return if its not */
-	    break;
-	}
-    }
-    while((nlHdr->nlmsg_seq != seqNum) || (nlHdr->nlmsg_pid != pId));
-
-    return msgLen;
-}
-
-int BibleSync::parseRoutes(struct nlmsghdr *nlHdr, struct route_info *rtInfo)
-{
-    struct rtmsg *rtMsg;
-    struct rtattr *rtAttr;
-    int rtLen;
-
-    rtMsg = (struct rtmsg *)NLMSG_DATA(nlHdr);
-
-    /* if route is not AF_INET or not in main table, return. */
-    if((rtMsg->rtm_family != AF_INET) || (rtMsg->rtm_table != RT_TABLE_MAIN))
-	return -1;
-
-    rtAttr = (struct rtattr *)RTM_RTA(rtMsg);
-    rtLen = RTM_PAYLOAD(nlHdr);
-
-    for(; RTA_OK(rtAttr,rtLen); rtAttr = RTA_NEXT(rtAttr,rtLen))
-    {
-	switch(rtAttr->rta_type)
-	{
-	case RTA_OIF:
-	    if_indextoname(*(int *)RTA_DATA(rtAttr), rtInfo->ifName);
-	    break;
-
-	case RTA_GATEWAY:
-	    memcpy(&rtInfo->gateWay, RTA_DATA(rtAttr), sizeof(rtInfo->gateWay));
-	    break;
-
-	case RTA_PREFSRC:
-	    memcpy(&rtInfo->srcAddr, RTA_DATA(rtAttr), sizeof(rtInfo->srcAddr));
-	    break;
-
-	case RTA_DST:
-	    memcpy(&rtInfo->dstAddr, RTA_DATA(rtAttr), sizeof(rtInfo->dstAddr));
-	    break;
-	}
-    }
-
-    return 0;
-}
-
-int BibleSync::get_default_if_name(char *name, socklen_t size)
-{
-    int found_default = 0;
-
-    struct nlmsghdr *nlMsg;
-    //struct rtmsg *rtMsg;
-    struct route_info route_info;
-    char msgBuf[4096];
-
-    int sock, len, msgSeq = 0;
-
-    name[1] = '\0';	// pre-set, in case of error.
-
-    if((sock = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE)) < 0)
+    if ((proc_route = fopen(PROC_ROUTE, "r")) == NULL)
     {
 	name[0] = 'x';
 	return -1;
     }
 
-    memset(msgBuf, 0, sizeof(msgBuf));
-
-    nlMsg = (struct nlmsghdr *)msgBuf;
-    //rtMsg = (struct rtmsg *)NLMSG_DATA(nlMsg);
-
-    nlMsg->nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
-    nlMsg->nlmsg_type = RTM_GETROUTE;
-
-    nlMsg->nlmsg_flags = NLM_F_DUMP | NLM_F_REQUEST;
-    nlMsg->nlmsg_seq = msgSeq++;
-    nlMsg->nlmsg_pid = getpid();
-
-    if(send(sock, nlMsg, nlMsg->nlmsg_len, 0) < 0)
+    while (fgets(line, 255, proc_route) != NULL)
     {
-	name[0] = 'y';
-	return -1;
-    }
+	if ((field = strchr(line, '\t')) == NULL)
+	    continue;			// invalid line?
 
-    if((len = readNlSock(sock, msgBuf, sizeof(msgBuf), msgSeq, getpid())) < 0)
-    {
-	name[0] = 'z';
-	return -1;
-    }
-
-    for(; NLMSG_OK(nlMsg,len); nlMsg = NLMSG_NEXT(nlMsg,len))
-    {
-	memset(&route_info, 0, sizeof(route_info));
-	if ( parseRoutes(nlMsg, &route_info) < 0 )
-	    continue;			// don't check: not set up
-
-	if (strstr((char *)inet_ntoa(route_info.dstAddr), "0.0.0.0"))
+	if (strncmp(field, "\t00000000\t", 10) == 0)
 	{
-	    // copy it over
-	    strcpy(name, route_info.ifName);
-	    found_default = 1;
+	    found = 1;
+	    *field = '\0';
+	    strcpy(name, line);
 	    break;
 	}
     }
+    fclose(proc_route);
 
-    close(sock);
-    return found_default;
+    if (!found)
+    {
+	// no default route?  fallback: we're holding a valid last line.
+	// so we arbitrarily choose whatever was found there.
+	*field = '\0';
+	strcpy(name, line);
+    }
+
+    return 0;
 }
 
-#include <netdb.h>
-#include <ifaddrs.h>
-
 void BibleSync::InterfaceAddress()
 {
     // cancel any old interface value.
@@ -1077,7 +1020,7 @@ void BibleSync::InterfaceAddress()
 
     char gw_if[IF_NAMESIZE];	// default gateway interface.
 
-    (void)get_default_if_name(gw_if, 100);
+    (void)get_default_if_name(gw_if);
 
     // if no error, search the interface list for that address.
     if (gw_if[0] != '\0')
@@ -1109,19 +1052,23 @@ void BibleSync::InterfaceAddress()
 
 // Solaris & BSD.
 
-//
 // this seeming grotesqueness is in fact the most general command
 // that could be found which finds the interface holding the default
 // route and then collects that interface's address.  handles both
-// solaris and bsd.  in fact, it suffices for linux as well, but
-// we're leaving the existing code above in place for linux, if for
-// no other reason than that it's more likely that someone will
-// foolishly screw up ifconfig output format in the linux world.
-//
-#define	ADDRESS	"PATH=/sbin:/usr/sbin:/bin:/usr/bin " \
-		"ifconfig \"`netstat -rn | egrep '^0\\.0\\.0\\.0|^default' | " \
-		"tr ' ' '\\n' | sed -e '/^$/d' | tail -1`\" | grep 'inet ' | " \
+// solaris and bsd.  also an instance for linux using "ip", in case
+// whoever builds this doesn't like depending on /proc/net/route.
+
+#ifdef linux
+#define	ADDRESS	"PATH=/sbin:/usr/sbin:/bin:/usr/bin "						\
+		"ip address show dev `ip route | egrep '^(default|0\\.0\\.0\\.0)' | "		\
+		"head -1 | sed 's/dev /DEV-/' | tr ' ' '\\n' | grep DEV | sed s/DEV-//` | "	\
+		"grep 'inet ' | tr '/ ' '\\n\\n' | grep '^[0-9.][0-9.]*' | head -1 | tr -d '\\n'"
+#else
+#define	ADDRESS	"PATH=/sbin:/usr/sbin:/bin:/usr/bin "				\
+		"ifconfig \"`netstat -rn | egrep '^0\\.0\\.0\\.0|^default' | "	\
+		"tr ' ' '\\n' | sed -e '/^$/d' | tail -1`\" | grep 'inet ' | "	\
 		"tr ' ' '\\n' | grep '^[0-9.][0-9.]*$' | head -1 | tr -d '\\n'"
+#endif
 
 void BibleSync::InterfaceAddress()
 {
